import requests
import re


class GenderResponse:
    """
        A class that acts as a wrapper for all responses which classify a name into a gender.

        Attributes:
            ID (str): The ID of the request
            first_name (str): The first name that was classified
            last_name (str): The last name that was classified
            likely_gender (str): The likely gender inferred from input (male or female)
            gender_scale (int): The gender scale inferred from input (-1 or 1)
            score (float): The score of the precision of the gender information provided ranging from 0 to 50
            probability_calibrated (float): The probability that the gender information provided is correct
    """

    ID = ""
    first_name = ""
    last_name = ""
    likely_gender = ""
    gender_scale = ""
    score = 0.0
    probability_calibrated = 0.0

    def __init__(self, api_response: dict):
        """ Constructor
        Args:
            api_response (dict): the json format (dict) of the NamsorAPI response received from a GET/POST request
        """

        self.ID = api_response['id']
        self.first_name = api_response['firstName']
        self.last_name = api_response['lastName']
        self.likely_gender = api_response['likelyGender']
        self.gender_scale = int(api_response['genderScale'])
        self.score = float(api_response['score'])
        self.probability_calibrated = float(
            api_response['probabilityCalibrated'])
        # Initializing the GenderResponse attributes with values of corresponding attribute names or keys of the API responsein json format


class GenderFullResponse:
    """
        A class that acts as a wrapper for all responses of POST requests which classify a full name into a gender.

        Attributes:
            ID (str): The ID of the request
            name (str): The name that was classified
            likely_gender (str): The likely gender inferred from input (male or female)
            gender_scale (int): The gender scale inferred from input (-1 or 1)
            score (float): The score of the precision of the gender information provided ranging from 0 to 50
    """

    ID = ""
    name = ""
    likely_gender = ""
    gender_scale = ""
    score = 0.0

    def __init__(self, api_response: dict):
        """ Constructor
        Args:
            api_response (dict): the json format (dict) of the NamsorAPI response received from a GET/POST request
        """

        self.ID = api_response['id']
        self.name = api_response['name']
        self.likely_gender = api_response['likelyGender']
        self.gender_scale = int(api_response['genderScale'])
        self.score = float(api_response['score'])


class OriginResponse:
    """
        A class that acts as a wrapper for all responses which infer the origin of an individual with a particular name.

        Attributes:
            ID (str): The ID of the request
            first_name (str): The first name that was classified
            last_name (str): The last name that was classified
            likely_gender (str): The likely gender inferred from input (male or female)
            gender_scale (int): The gender scale inferred from input (-1 or 1)
            score (float): The score of the precision of the gender information provided ranging from 0 to 50
            country_origin (str): The most likely country that the individual with this name belongs to.
            country_origin_alt (str): A potential alternative country that the individual with this name belongs to.
            probability_calibrated (float): The probability that the classification is correct
            region_origin (str):
            top_region_origin (str):
            sub_region_origin (str): 
    """
    ID = ""
    first_name = ""
    last_name = ""
    likely_gender = ""
    score = 0.0
    # Maybe use the country codes class here?
    country_origin = ""
    country_origin_alt = ""
    region_origin = ""
    top_region_origin = ""
    sub_region_origin = ""

    def __init__(self, api_response: dict):
        """ Constructor
        Args:
            api_response (dict): the json format (dict) of the NamsorAPI response received from a GET/POST request
        """
        self.ID = api_response['id']
        self.first_name = api_response['firstName']
        self.last_name = api_response['lastName']
        self.country_origin = api_response['countryOrigin']
        self.country_origin_alt = api_response['countryOriginAlt']
        self.score = float(api_response['score'])
        self.region_origin = api_response['regionOrigin']
        self.top_region_origin = api_response['topRegionOrigin']
        self.sub_region_origin = api_response['subRegionOrigin']


class CountryResponse:

    def __init__(self, api_response: dict):
        """ Constructor

        Args:
            api_response (dict): the json format (dict) of the NamsorAPI response received from a GET/POST request
        """

        self.ID = api_response['id']
        self.name = api_response['name']
        self.score = float(api_response['score'])
        self.country = api_response['country']
        self.country_alt = api_response['countryAlt']
        self.region = api_response['region']
        self.top_region = api_response['topRegion']
        self.sub_region = api_response['subRegion']


class RaceEthnicityResponse:
    """
        A class that acts as a wrapper for all responses which infer the ethnicity of an individual with a particular name.

        Attributes:
            ID (str): The ID of the request
            first_name (str): The first name that was classified
            last_name (str): The last name that was classified
            score (float): The score of the precision of the gender information provided ranging from 0 to 50
            race_ethnicity (str): The most likely ethnicity of an individual with said name.
            race_ethnicity_alt (str): A potential alternative ethnicity that the individual can have.
    """
    ID = ""
    first_name = ""
    last_name = ""
    race_ethnicity_alt = "W_NL"
    race_ethnicity = "W_NL"
    score = 0

    def __init__(self, api_response: dict):
        """ Constructor
        Args:
            api_response (dict): the json format (dict) of the NamsorAPI response received from a GET/POST request
        """
        self.ID = api_response['id']
        self.first_name = api_response['firstName']
        self.last_name = api_response['lastName']
        self.race_ethnicity = api_response['raceEthnicity']
        self.race_ethnicity_alt = api_response['raceEthnicityAlt']
        self.score = api_response['score']


class DiasporaResponse:
    """
        A class that acts as a wrapper for the diaspora response object.

        Attributes:
            ID (str): The ID of the request
            first_name (str): The first name that was classified
            last_name (str): The last name that was classified
            score (float): The score of the precision of the gender information provided ranging from 0 to 50
            race_ethnicity (str): The most likely ethnicity of an individual with said name.
            race_ethnicity_alt (str): A potential alternative ethnicity that the individual can have.
            lifted (str):
            country (str): The ISO2 code of the country to which an indiviudal with this name liekly belongs to.
    """

    ID = ""
    first_name = ""
    last_name = ""
    score = 0.0
    ethnicity_alt = ""
    ethnicity = ""
    lifted = ""
    country = ""

    def __init__(self, api_response: dict):
        """ Constructor
        Args:
            api_response (dict): the json format (dict) of the NamsorAPI response received from a GET/POST request
        """
        self.ID = api_response['id']
        self.first_name = api_response['firstName']
        self.last_name = api_response['lastName']
        self.score = float(api_response['score'])
        self.ethnicity_alt = api_response['ethnicityAlt']
        self.ethnicity = api_response['ethnicity']
        self.lifted = api_response['lifted']
        self.country = api_response['countryIso2']


class ParseNameResponse:
    """
        A class that acts as a wrapper for the response object of all the parseName routes.

        Attributes:
            ID (str): The ID of the request
            name (str): The full name to be parsed
            name_parser_type (NameParserTypeWrapper): An object which contains data about how many words from the full name were classified into first names and last names.

            name_parser_type_alt (NameParserTypeWrapper): An object which contains another potential form of data about how many words from the full name were classified into first names and last names.

            first_last_name (FirstLastNameWrapper): An object which contains the most likely first name/ last name format that the full name would be arranged in.

            score (float): The score of the precision of the gender information provided ranging from 0 to 50
    """
    ID = ""
    name = ""
    name_parser_type = ""
    name_parser_type_alt = ""
    first_last_name = ""
    score = ""

    def __init__(self, api_response: dict):
        """ Constructor
        Args:
            api_response (dict): the json format (dict) of the NamsorAPI response received from a GET/POST request
        """
        self.ID = api_response['id']
        self.name = api_response['name']
        self.name_parser_type = NameParserTypeWrapper(
            api_response['nameParserType'])
        self.name_parser_type_alt = NameParserTypeWrapper(
            api_response['nameParserTypeAlt'])
        self.first_last_name = FirstLastNameWrapper(
            api_response['firstLastName'])
        self.score = api_response['score']


class NameParserTypeWrapper:
    """A class which acts as a wrapper for the nameParserType key in the parseName response objects.

    Attributes:
        first_name_count (int): The number of words classified as first names.
        last_name_count (int): The number of words classified as last names.

    """
    _raw_string = ""
    first_name_count = 0
    last_name_count = 0

    def __init__(self, raw_string: str):
        """ Constructor
        Args:
            raw_string (str): The raw string in the response object representing the value of the 'nameParserType' key.
        """
        self.raw_string = raw_string

        if raw_string != None and raw_string != "null":
            regex = ""
            if raw_string.startswith("LN"):
                regex = re.fullmatch(r"LN([0-9]+)FN([0-9]+)", raw_string)
            else:
                regex = re.fullmatch(r"FN([0-9]+)LN([0-9]+)", raw_string)
            self.first_name_count = int(regex.groups()[0])
            self.last_name_count = int(regex.groups()[1])
            print(self)

    def __repr__(self):
        return self.raw_string

    def __eq__(self, other):
        return self.first_name_count == other.first_name_count \
            and self.last_name_count == other.last_name_count \
            and self.raw_string == str(other)


class FirstLastNameWrapper:
    """A class which acts as a wrapper for the value of the 'firstLastName' key in the parseName response objects.

    Attributes:
        first_name (str): The portion of the full name classified as the first name.
        last_name_count (int): The portion of the full name classified as the last name.

    """

    ID = ""
    first_name = ""
    last_name = ""

    def __init__(self, raw_dict: dict):
        """ Constructor
        Args:
            raw_dict (dict): The raw json object in the response object representing the value of the 'firstLastName' key. 
        """
        self.ID = raw_dict['id']
        self.first_name = raw_dict['firstName']
        self.last_name = raw_dict['lastName']

    def __repr__(self):
        return f'First Name: {self.first_name} | Last Name: {self.last_name}'

    def __eq__(self, other):
        return self.first_name == other.first_name \
            and self.last_name == other.last_name \
            and self.ID == other.ID



Sample_Responses = {
    #        ------------------------------------------------------------------------------------
    #                                   GenderResponse Samples
    #        ------------------------------------------------------------------------------------
    "GenderResponses":
        [{
            'id': '790',
            'firstName': 'Cynthia',
            'lastName': 'Martin',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 38.93331818668968,
            'probabilityCalibrated': 0.992457769300861
        }, {
            'id': '483',
            'firstName': 'Kelly',
            'lastName': 'Herrera',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 13.206913643992483,
            'probabilityCalibrated': 0.7568709219543044
        }, {
            'id': '738',
            'firstName': 'Margaret',
            'lastName': 'Phillips',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 28.28155981127263,
            'probabilityCalibrated': 0.9833262796012671
        }, {
            'id': '804',
            'firstName': 'Kevin',
            'lastName': 'Armstrong',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 36.82855624853989,
            'probabilityCalibrated': 0.9911835223666028
        }, {
            'id': '696',
            'firstName': 'Alicia',
            'lastName': 'Elliott',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 40.84375144891046,
            'probabilityCalibrated': 0.993543334003477
        }, {
            'id': '573',
            'firstName': 'David',
            'lastName': 'Perez',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 42.35570879385797,
            'probabilityCalibrated': 0.9936257852306357
        }, {
            'id': '31',
            'firstName': 'Breanna',
            'lastName': 'Lang',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 27.28303588458596,
            'probabilityCalibrated': 0.9813295829845597
        }, {
            'id': '388',
            'firstName': 'Mary',
            'lastName': 'Stein',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 20.892064581104407,
            'probabilityCalibrated': 0.9418945605393836
        }, {
            'id': '307',
            'firstName': 'Tonya',
            'lastName': 'Bright',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 21.72395268910261,
            'probabilityCalibrated': 0.9545684450756573
        }, {
            'id': '59',
            'firstName': 'Shelly',
            'lastName': 'Novak',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 18.845089011447104,
            'probabilityCalibrated': 0.9107087174326258
        }],
    #        ------------------------------------------------------------------------------------
    #                                   GenderFullResponse Samples
    #        ------------------------------------------------------------------------------------
    "GenderFullResponses": [{
            'id': '113',
            'name': 'Brett Petersen',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 33.691152352881794
        }, {
            'id': '195',
            'name': 'April Jackson DVM',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 9.315183448265584
        }, {
            'id': '498',
            'name': 'Tyrone Sanders',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 35.73948515061509
        }, {
            'id': '506',
            'name': 'Michael Green',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 30.963259191770305
        }, {
            'id': '914',
            'name': 'Mark Smith',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 44.99008221179214
        }, {
            'id': '522',
            'name': 'Angela Flores',
            'likelyGender': 'female',
            'genderScale': 1.0,
            'score': 24.634573447725124
        }, {
            'id': '563',
            'name': 'Richard Gonzalez',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 52.82109285593519
        }, {
            'id': '634',
            'name': 'Brandon Rogers',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 39.072922098778506
        }, {
            'id': '625',
            'name': 'John Clark',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 55.9746072512175
        }, {
            'id': '609',
            'name': 'Zachary Jacobs',
            'likelyGender': 'male',
            'genderScale': -1.0,
            'score': 46.82023753356409
        }],
    #        ------------------------------------------------------------------------------------
    #                                   OriginResponse Samples
    #        ------------------------------------------------------------------------------------
    "OriginResponses": [{
            'id': '31',
            'firstName': 'Stephen',
            'lastName': 'Coleman',
            'countryOrigin': 'GB',
            'countryOriginAlt': 'IE',
            'score': 1.8947393406337736,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Northern Europe'
        }, {
            'id': '146',
            'firstName': 'Michael',
            'lastName': 'Burke',
            'countryOrigin': 'IE',
            'countryOriginAlt': 'GB',
            'score': 10.494827631725524,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Northern Europe'
        }, {
            'id': '412',
            'firstName': 'Adrian',
            'lastName': 'Church',
            'countryOrigin': 'GB',
            'countryOriginAlt': 'CH',
            'score': 12.40215333766311,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Northern Europe'
        }, {
            'id': '454',
            'firstName': 'Austin',
            'lastName': 'Boyd',
            'countryOrigin': 'GB',
            'countryOriginAlt': 'IE',
            'score': 6.311295115561374,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Northern Europe'
        }, {
            'id': '428',
            'firstName': 'Michelle',
            'lastName': 'Trevino',
            'countryOrigin': 'FR',
            'countryOriginAlt': 'IT',
            'score': 0.34130148343600725,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Western Europe'
        }, {
            'id': '537',
            'firstName': 'Jordan',
            'lastName': 'Gonzalez',
            'countryOrigin': 'ES',
            'countryOriginAlt': 'FR',
            'score': 21.71707930344048,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Southern Europe'
        }, {
            'id': '543',
            'firstName': 'Matthew',
            'lastName': 'Garcia',
            'countryOrigin': 'ES',
            'countryOriginAlt': 'GB',
            'score': 10.620154841037358,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Southern Europe'
        }, {
            'id': '876',
            'firstName': 'Beth',
            'lastName': 'Evans',
            'countryOrigin': 'GB',
            'countryOriginAlt': 'IE',
            'score': 14.819170193754193,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Northern Europe'
        }, {
            'id': '978',
            'firstName': 'Elizabeth',
            'lastName': 'Miller',
            'countryOrigin': 'GB',
            'countryOriginAlt': 'IE',
            'score': 6.241282311663333,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Northern Europe'
        }, {
            'id': '680',
            'firstName': 'Alexis',
            'lastName': 'Moreno',
            'countryOrigin': 'ES',
            'countryOriginAlt': 'FR',
            'score': 5.7003339923628245,
            'regionOrigin': 'Europe',
            'topRegionOrigin': 'Europe',
            'subRegionOrigin': 'Southern Europe'
        }],
    #        ------------------------------------------------------------------------------------
    #                                   CountryResponse Samples
    #        ------------------------------------------------------------------------------------
    "CountryResponses": [{
            'id': '576',
            'name': 'Jamie Fleming',
            'score': 7.06264456142918,
            'country': 'GB',
            'countryAlt': 'IE',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Northern Europe'
        }, {
            'id': '212',
            'name': 'John Roberts',
            'score': 3.0803229164056143,
            'country': 'GB',
            'countryAlt': 'IE',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Northern Europe'
        }, {
            'id': '573',
            'name': 'Jack Rhodes',
            'score': 13.510695042355879,
            'country': 'GB',
            'countryAlt': 'IE',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Northern Europe'
        }, {
            'id': '198',
            'name': 'Amanda Mcclure',
            'score': 4.293304555301407,
            'country': 'GB',
            'countryAlt': 'IE',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Northern Europe'
        }, {
            'id': '407',
            'name': 'James Wood',
            'score': 10.623796000676329,
            'country': 'GB',
            'countryAlt': 'IE',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Northern Europe'
        }, {
            'id': '267',
            'name': 'Bryan Griffith',
            'score': 1.7971469008117864,
            'country': 'GB',
            'countryAlt': 'IE',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Northern Europe'
        }, {
            'id': '200',
            'name': 'Karen Herrera',
            'score': 12.10225278411596,
            'country': 'ES',
            'countryAlt': 'GB',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Southern Europe'
        }, {
            'id': '188',
            'name': 'Darlene Proctor',
            'score': 9.608991621223403,
            'country': 'GB',
            'countryAlt': 'IE',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Northern Europe'
        }, {
            'id': '105',
            'name': 'Donald Miller',
            'score': 9.776572059793308,
            'country': 'GB',
            'countryAlt': 'AT',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Northern Europe'
        }, {
            'id': '109',
            'name': 'Sabrina Lara',
            'score': 8.45910530844296,
            'country': 'ES',
            'countryAlt': 'IT',
            'region': 'Europe',
            'topRegion': 'Europe',
            'subRegion': 'Southern Europe'
        }],
    #        ------------------------------------------------------------------------------------
    #                                   RaceEthnicityResponse Samples
    #        ------------------------------------------------------------------------------------
    "RaceEthnicityResponses": [{
            'id': '607',
            'firstName': 'Patrick',
            'lastName': 'Rasmussen',
            'raceEthnicityAlt': 'A',
            'raceEthnicity': 'W_NL',
            'score': 22.948248540996016
        }, {
            'id': '282',
            'firstName': 'Christopher',
            'lastName': 'Kelly',
            'raceEthnicityAlt': 'B_NL',
            'raceEthnicity': 'W_NL',
            'score': 7.955888738554523
        }, {
            'id': '914',
            'firstName': 'Thomas',
            'lastName': 'Mcdonald',
            'raceEthnicityAlt': 'B_NL',
            'raceEthnicity': 'W_NL',
            'score': 10.201947264489698
        }, {
            'id': '241',
            'firstName': 'Cameron',
            'lastName': 'Graham',
            'raceEthnicityAlt': 'W_NL',
            'raceEthnicity': 'B_NL',
            'score': 3.5776916899124154
        }, {
            'id': '200',
            'firstName': 'Tracy',
            'lastName': 'Smith',
            'raceEthnicityAlt': 'W_NL',
            'raceEthnicity': 'B_NL',
            'score': 0.4432138744951581
        }, {
            'id': '908',
            'firstName': 'Kimberly',
            'lastName': 'Sellers',
            'raceEthnicityAlt': 'B_NL',
            'raceEthnicity': 'W_NL',
            'score': 4.316583054898527
        }, {
            'id': '37',
            'firstName': 'Eric',
            'lastName': 'Wyatt',
            'raceEthnicityAlt': 'B_NL',
            'raceEthnicity': 'W_NL',
            'score': 4.876180530048689
        }, {
            'id': '606',
            'firstName': 'Joann',
            'lastName': 'Hunter',
            'raceEthnicityAlt': 'B_NL',
            'raceEthnicity': 'W_NL',
            'score': 5.033235484800428
        }, {
            'id': '941',
            'firstName': 'Rebecca',
            'lastName': 'Alexander',
            'raceEthnicityAlt': 'B_NL',
            'raceEthnicity': 'W_NL',
            'score': 6.45911528690918
        }, {
            'id': '58',
            'firstName': 'Jennifer',
            'lastName': 'Martinez',
            'raceEthnicityAlt': 'W_NL',
            'raceEthnicity': 'HL',
            'score': 28.71009666363218
        }],
        #        ------------------------------------------------------------------------------------
        #                                   DiasporaResponse Samples
        #        ------------------------------------------------------------------------------------
        "DiasporaResponses": [{
            'id': '538',
            'firstName': 'Charles',
            'lastName': 'Juarez',
            'score': 5.018694044348786,
            'ethnicityAlt': 'Portuguese',
            'ethnicity': 'Hispanic',
            'lifted': True,
            'countryIso2': 'LY'
        }, {
            'id': '50',
            'firstName': 'Craig',
            'lastName': 'Rubio',
            'score': 1.6060483558594125,
            'ethnicityAlt': 'Portuguese',
            'ethnicity': 'Hispanic',
            'lifted': True,
            'countryIso2': 'GF'
        }, {
            'id': '518',
            'firstName': 'Ashley',
            'lastName': 'Johnson',
            'score': 11.337119830452618,
            'ethnicityAlt': 'British',
            'ethnicity': 'Liberia',
            'lifted': True,
            'countryIso2': 'MK'
        }, {
            'id': '139',
            'firstName': 'Jeffrey',
            'lastName': 'Murphy',
            'score': 16.499316179724513,
            'ethnicityAlt': 'British',
            'ethnicity': 'Irish',
            'lifted': False,
            'countryIso2': 'TL'
        }, {
            'id': '564',
            'firstName': 'Darrell',
            'lastName': 'Perez',
            'score': 11.129313165321383,
            'ethnicityAlt': 'Portuguese',
            'ethnicity': 'Hispanic',
            'lifted': True,
            'countryIso2': 'TO'
        }, {
            'id': '458',
            'firstName': 'Jeremiah',
            'lastName': 'Allison',
            'score': 16.470926891304973,
            'ethnicityAlt': 'Irish',
            'ethnicity': 'British',
            'lifted': False,
            'countryIso2': 'NP'
        }, {
            'id': '750',
            'firstName': 'Jennifer',
            'lastName': 'Moore',
            'score': 10.110044737078029,
            'ethnicityAlt': 'Irish',
            'ethnicity': 'British',
            'lifted': False,
            'countryIso2': 'PG'
        }, {
            'id': '985',
            'firstName': 'Adam',
            'lastName': 'Phillips',
            'score': 5.448428097566382,
            'ethnicityAlt': 'Jewish',
            'ethnicity': 'British',
            'lifted': False,
            'countryIso2': 'LB'
        }, {
            'id': '705',
            'firstName': 'Robert',
            'lastName': 'Davies',
            'score': 13.48239707162697,
            'ethnicityAlt': 'Jewish',
            'ethnicity': 'British',
            'lifted': False,
            'countryIso2': 'SN'
        }, {
            'id': '340',
            'firstName': 'Kevin',
            'lastName': 'Kline',
            'score': 5.9249709299274675,
            'ethnicityAlt': 'French',
            'ethnicity': 'Jewish',
            'lifted': False,
            'countryIso2': 'TF'
        }],
    #        ------------------------------------------------------------------------------------
    #                                   ParsedNameResponse Samples
    #        ------------------------------------------------------------------------------------
    "ParsedNameResponses": [{
        'id': '650',
        'name': 'Oscar Lawrence',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'Oscar',
            'lastName': 'Lawrence'
        },
        'score': 1.1028782830206576
    }, {
        'id': '304',
        'name': 'Jesse Johnston',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'Jesse',
            'lastName': 'Johnston'
        },
        'score': 23.63613617145121
    }, {
        'id': '802',
        'name': 'Kenneth Mcgee',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'Kenneth',
            'lastName': 'Mcgee'
        },
        'score': 55.37036489790589
    }, {
        'id': '618',
        'name': 'Lucas Torres',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'Lucas',
            'lastName': 'Torres'
        },
        'score': 27.728602874596294
    }, {
        'id': '889',
        'name': 'Rhonda Herrera',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'Rhonda',
            'lastName': 'Herrera'
        },
        'score': 37.22560301808376
    }, {
        'id': '959',
        'name': 'Molly Dixon',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'Molly',
            'lastName': 'Dixon'
        },
        'score': 15.556550881794742
    }, {
        'id': '974',
        'name': 'Devin Fernandez',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'Devin',
            'lastName': 'Fernandez'
        },
        'score': 37.21015080658581
    }, {
        'id': '254',
        'name': 'Stacey George',
        'nameParserType': 'LN1FN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'George',
            'lastName': 'Stacey'
        },
        'score': 9.445291910747269
    }, {
        'id': '205',
        'name': 'David Davis',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'David',
            'lastName': 'Davis'
        },
        'score': 15.079782026578242
    }, {
        'id': '434',
        'name': 'Kevin Garza',
        'nameParserType': 'FN1LN1',
        'nameParserTypeAlt': None,
        'firstLastName': {
            'id': None,
            'firstName': 'Kevin',
            'lastName': 'Garza'
        },
        'score': 46.719939723728054
    }]
}



import math
import time
import xlsxwriter
import re
from abc import ABC, abstractmethod
import requests

import faker
import random

from namsorclient.models import *
from namsorclient.country_codes import CountryCodes


class BatchItem(ABC):
    """ An abstract class representing a general batch item.
    """
    pass


class Batch(ABC):
    """ An abstract class representing a general batch in which items can be added.
    """

    response = []

    @abstractmethod
    def batch_item_converter(self):
        pass

    @abstractmethod
    def addItem(self, item: BatchItem):
        """ An abstract method that appends an item of type BatchItem (any subclass of BatchItem) to items, a list.
        Args:
            item (BatchItem): The item of type BatchItem (any subclass of BatchItem) which is to be 
            appended to items, a list.
        """
        self.items.append(item)

    def api_post(self, url: str, data: dict, api_key: str) -> requests.models.Response:
        """ Returns a response containing desired information of the data.
        Args:
            url (str): Ending portion of NamsorAPI url to desired section.
            data (dict): The high throughput data to process.
        Returns:
            requests.models.Response: The response received from this POST request.
        """
        print("FUCKKKKKKKKKKK")
        BASE_URL = "https://v2.namsor.com/NamSorAPIv2/api2/json/"
        r = requests.post(url=f"{BASE_URL}{url}", headers={"X-API-KEY":
                                                           api_key}, json=data)
        if r.status_code == 401:
            # The client has not entered his/her API key or the API key is incorrect.
            raise Exception('Invalid API Key')
        elif r.status_code == 403:
            # The client's amount of requests has reached the maximum amount he/she can send in a month or his/her subscription plan has been cancelled.
            raise Exception("API Limit Reached or API Key Disabled")
        return r

    def classify(self, api_key: str) -> list:
        """ Takes in the user's API key and returns responses of this batch's request(s) and response type in the form of a list.

        Args:
            api_key (str): The user's API key.

        Returns:
            list: A list of responses of this batch's response type.
        """

        # Stores the batch's items' data in appropriate JSON format in personal_names_list
        personal_names_list = self.batch_item_converter()
        
        response_list = []
        # Data is separated into blocks to bypass the 100 item limit
        item_list = list_separator(personal_names_list)

        for item in item_list:
            # Data is to be put in the appropriate format and be passed in as an argument of the POST request
            payload = {}
            payload['personalNames'] = item
            response = self.api_post(url=self.url, data=payload, api_key=api_key).json()[
                'personalNames']

            # This response from the POST request is appended to the previous responses
            self.response += response
            print(response)
            for i in range(len(response)):
                response_list.append(self.response_type(response[i]))
            # A one second delay for latency
            time.sleep(1)

        return response_list

    def export_to_excel(self, file_name: str):
        """ Creates/overwrites an excel file and represents the batch's items' data in a spreadsheet form.

        Args:
            file_name (str): The desired Excel file name.
        """
        workbook = xlsxwriter.Workbook(file_name)
        worksheet = workbook.add_worksheet()

        for column, item in enumerate(self.response[0].keys()):

            # Regex to convert the keys from camel-case to Capital Case separated by hyphens
            # Eg: firstName to First-Name
            match = re.match(
                r'([a-z]+)([A-Z]*[a-z]*)([A-Z]*[a-z]*)([A-Z]*[a-z]*)', item).groups()
            row_title = '-'.join([i[0].upper() + i[1:]
                                  for i in match if i != ''])
            worksheet.write(0, column, row_title)

        for num, item in enumerate(self.response):
            for column, value in enumerate(item.values()):
                worksheet.write(num+1, column, str(value))

        workbook.close()


class GenderBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely gender of multiple names, detecting automatically the cultural context.
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class GenderBatchItem(BatchItem):
        """ 
            A class representing an item of a GenderBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
        """

        def __init__(self, first_name: str, last_name: str, ID="unassigned"):
            """ Constructor

            Args:
                first_name (str): The desired first name.
                last_name (str): The desired last name.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.first_name = first_name
            self.last_name = last_name

    url = "genderBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = GenderResponse

    def addItem(self, first_name: str, last_name: str, ID="unassigned"):
        """ Adds a GenderBatchItem, with the required input, to the batch.
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.GenderBatchItem(first_name, last_name, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "firstName": item.first_name,
                "lastName": item.last_name,
            })

        return items_list


class GenderGeoBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely gender of multiple names, each given a local context (ISO2 country code).
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class GenderGeoBatchItem(BatchItem):
        """ 
            A class representing an item of a GenderGeoBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
               country_code (CountryCodes): The desired country code.
        """

        def __init__(self, first_name: str, last_name: str, country_code: CountryCodes, ID="unassigned"):
            """Constructor
            Args:
                first_name (str): The desired first name.
                last_name (str): The desired last name.
                country_code (CountryCodes): The desired country code.
                ID (str, optional): the ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.first_name = first_name
            self.last_name = last_name
            self.country_code = country_code.value

    url = "genderGeoBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = GenderResponse

    def addItem(self, first_name: str, last_name: str, country_code: CountryCodes, ID="unassigned"):
        """ Adds a GenderGeoBatchItem, with the required input, to the batch.
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            country_code (CountryCodes): The desired country code.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.GenderGeoBatchItem(first_name, last_name, country_code, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "firstName": item.first_name,
                "lastName": item.last_name,
                "countryIso2": item.country_code
            })

        return items_list


class ParsedGenderBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely gender of multiple fully parsed names, detecting automatically the cultural context.
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class ParsedGenderBatchItem(BatchItem):
        """ 
            A class representing an item of a ParsedGenderBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
               prefix_or_title (str): The desired prefix or title.
               suffix (str): The desired suffix.
               middle_name (str): The desired middle name.
        """

        def __init__(self, first_name: str, last_name: str, prefix_or_title: str, suffix: str, middle_name: str, ID="unassigned"):
            self.ID = ID
            self.first_name = first_name
            self.last_name = last_name
            self.prefix_or_title = prefix_or_title
            self.suffix = suffix
            self.middle_name = middle_name

    url = "parsedGenderBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = GenderResponse

    def addItem(self, first_name: str, last_name: str, prefix_or_title: str, suffix: str, middle_name: str, ID="unassigned"):
        """ Adds a ParsedGenderBatchItem, with the required input, to the batch.
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            prefix_or_title (str): The desired prefix or title
            suffix (str): The desired suffix
            middle_name (str): The desired middle name
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.ParsedGenderBatchItem(first_name,
                                                   last_name, prefix_or_title, suffix, middle_name, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "firstName": item.first_name,
                "lastName": item.last_name,
                "prefixOrTitle": item.prefix_or_title,
                "suffix": item.suffix,
                "middleName": item.middle_name,
            })

        return items_list


class ParsedGenderGeoBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely gender of multiple fully parsed names, detecting automatically the cultural context.
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class ParsedGenderGeoBatchItem(BatchItem):
        """ 
            A class representing an item of a ParsedGenderGeoBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
               prefix_or_title (str): The desired prefix or title
               suffix (str): The desired suffix
               middle_name (str): The desired middle name
               country_code (CountryCodes): The desired country code.
        """

        def __init__(self, first_name: str, last_name: str, prefix_or_title: str, suffix: str, middle_name: str, country_code: CountryCodes, ID="unassigned"):
            """Constructor
            Args:
                first_name (str): The desired first name.
                last_name (str): The desired last name.
                prefix_or_title (str): The desired prefix or title
                suffix (str): The desired suffix
                middle_name (str): The desired middle name
                country_code (CountryCodes): The desired country code.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.first_name = first_name
            self.last_name = last_name
            self.prefix_or_title = prefix_or_title
            self.suffix = suffix
            self.middle_name = middle_name
            self.country_code = country_code.value

    url = "parsedGenderGeoBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = GenderResponse

    def addItem(self, first_name: str, last_name: str, prefix_or_title: str, suffix: str, middle_name: str, country_code: CountryCodes, ID="unassigned"):
        """ Adds a ParsedGenderGeoBatchItem, with the required input, to the batch.
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            prefix_or_title (str): The desired prefix or title
            suffix (str): The desired suffix
            middle_name (str): The desired middle name
            country_code (CountryCodes): The desired country code.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.ParsedGenderGeoBatchItem(first_name, last_name,
                                                      prefix_or_title, suffix, middle_name, country_code, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "firstName": item.first_name,
                "lastName": item.last_name,
                "prefixOrTitle": item.prefix_or_title,
                "suffix": item.suffix,
                "middleName": item.middle_name,
                "countryIso2": item.country_code
            })

        return items_list


class GenderFullBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely gender of multiple full names, detecting automatically the cultural context.
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class GenderFullBatchItem(BatchItem):
        """ 
            A class representing an item of a GenderFullBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               name (str): The desired name.
        """

        def __init__(self, name: str, ID="unassigned"):
            """Constructor
            Args:
                name (str): The desired name.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.name = name

    url = "genderFullBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = GenderFullResponse

    def addItem(self, name: str, ID="unassigned"):
        """ Adds a GenderFullBatchItem, with the required input, to the batch.
        Args:
            name (str): The desired name.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.GenderFullBatchItem(name, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "name": item.name,
            })

        return items_list


class GenderFullGeoBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely gender of multiple full names, with a given cultural context (country ISO2 code).
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class GenderFullGeoBatchItem(BatchItem):
        """ 
            A class representing an item of a GenderFullGeoBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               name (str): The desired name.
               country_code (CountryCodes): The desired country code.
        """

        def __init__(self, name: str,  country_code: CountryCodes,  ID="unassigned"):
            """ Constructor
            Args:
                name (str): The desired name.
                country_code (CountryCodes): The desired country code.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.name = name
            self.country_code = country_code.value

    url = "genderFullGeoBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = GenderFullResponse

    def addItem(self, name: str,  country_code: CountryCodes,  ID="unassigned"):
        """ Adds a GenderFullGeoBatchItem, with the required input, to the batch.
        Args:
            name (str): The desired name.
            country_code (CountryCodes): The desired country code.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.GenderFullGeoBatchItem(name, country_code, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "name": item.name,
                "countryIso2": item.country_code,
            })

        return items_list


class OriginBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely country of origin of multiple names, detecting automatically the cultural context.
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class OriginBatchItem(BatchItem):
        """ 
            A class representing an item of an OriginBatchItem object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
        """

        def __init__(self, first_name: str, last_name: str,  ID="unassigned"):
            """Constructor
            Args:
                first_name (str): The desired first name.
                last_name (str): The desired last name.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.first_name = first_name
            self.last_name = last_name

    url = "originBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = OriginResponse

    def addItem(self, first_name: str, last_name: str,  ID="unassigned"):
        """ Adds a OriginBatchItem, with the required input, to the batch.
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.OriginBatchItem(first_name, last_name, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "firstName": item.first_name,
                "lastName": item.last_name,
            })

        return items_list


class CountryBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely country of residence of multiple personal full names, or surnames. Assumes names as they are in the country of residence OR the country of origin.
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class CountryBatchItem(BatchItem):
        """ 
            A class representing an item of an OriginBatchItem object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
        """

        def __init__(self, name: str,  ID="unassigned"):
            """Constructor
            Args:
                name (str): The desired name
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.name = name

    url = "countryBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = CountryResponse

    def addItem(self, name: str,  ID="unassigned"):
        """ Adds a CountryBatchItem, with the required input, to the batch. 
        Args:
            name (str): The desired name
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.CountryBatchItem(name, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "name": item.name,
            })

        return items_list


class US_RaceEthnicityBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer a US resident's likely race/ethnicity according to US Census taxonomy W_NL (white, non latino), HL (hispano latino),  A (asian, non latino), B_NL (black, non latino).
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class US_RaceEthnicityBatchItem(BatchItem):
        """ 
            A class representing an item of a US_RaceEthnicityBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
               country_code (CountryCodes): The desired country code.
        """

        def __init__(self, first_name: str, last_name: str,  country_code: CountryCodes,   ID="unassigned"):
            """Constructor
            Args:
                first_name (str): The desired first name.
                last_name (str): The desired last name.
                country_code (CountryCodes): The desired country code.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.first_name = first_name
            self.last_name = last_name
            self.country_code = country_code.value

    url = "usRaceEthnicityBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = RaceEthnicityResponse

    def addItem(self, first_name: str, last_name: str,  country_code: CountryCodes,   ID="unassigned"):
        """ Adds a US_RaceEthnicityBatchItem, with the required input, to the batch. 
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            country_code (CountryCodes): The desired country code.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.US_RaceEthnicityBatchItem(
            first_name, last_name, country_code, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "firstName": item.first_name,
                "lastName": item.last_name,
                "countryIso2": item.country_code, })

        return items_list


class US_ZipRaceEthnicityBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer a US resident's likely race/ethnicity according to US Census taxonomy, using ZIP5 code info. Output is W_NL (white, non latino), HL (hispano latino),  A (asian, non latino), B_NL (black, non latino).
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class US_ZipRaceEthnicityBatchItem(BatchItem):
        """ 
            A class representing an item of a US_ZipRaceEthnicityBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
               country_code (CountryCodes): The desired country code.
               zip_code (str): The desired zip code.
        """

        def __init__(self, first_name: str, last_name: str,  country_code: CountryCodes, zip_code: str,  ID="unassigned"):
            """Constructor
            Args:
                first_name (str): The desired first name.
                last_name (str): The desired last name.
                country_code (CountryCodes): The desired country code.
                zip_code (str): The desired zip code.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.first_name = first_name
            self.last_name = last_name
            self.country_code = country_code.value
            self.zip_code = zip_code

    url = "usZipRaceEthnicityBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = RaceEthnicityResponse

    def addItem(self, first_name: str, last_name: str,  country_code: CountryCodes, zip_code: str, ID="unassigned"):
        """ Adds a CountryBatchItem, with the required input, to the batch. 
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            country_code (CountryCodes): The desired country code.
            zip_code (str): The desired zip code.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.US_ZipRaceEthnicityBatchItem(
            first_name, last_name, country_code, zip_code, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "firstName": item.first_name,
                "lastName": item.last_name,
                "countryIso2": item.country_code,
                "zipCode": item.zip_code, })

        return items_list


class DiasporaBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely ethnicity/diaspora of up to 100 personal names, given a country of residence ISO2 code (ex. US, CA, AU, NZ etc.)
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class DiasporaBatchItem(BatchItem):
        """ 
            A class representing an item of a DiasporaBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               first_name (str): The desired first name.
               last_name (str): The desired last name.
               country_code (CountryCodes): The desired country code.
        """

        def __init__(self, first_name: str, last_name: str,  country_code: CountryCodes,   ID="unassigned"):
            """Constructor
            Args:
                first_name (str): The desired first name.
                last_name (str): The desired last name.
                country_code (CountryCodes): The desired country code.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.first_name = first_name
            self.last_name = last_name
            self.country_code = country_code.value

    url = "diasporaBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = DiasporaResponse

    def addItem(self, first_name: str, last_name: str,  country_code: CountryCodes,   ID="unassigned"):
        """ Adds a DiasporaBatchItem, with the required input, to the batch. 
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            country_code (CountryCodes): The desired country code.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.DiasporaBatchItem(first_name, last_name, country_code, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "firstName": item.first_name,
                "lastName": item.last_name,
                "countryIso2": item.country_code, })

        return items_list


class ParseNameBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely first/last name structure of a name, ex. John Smith or SMITH, John or SMITH; John.
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class ParseNameBatchItem(BatchItem):
        """ 
            A class representing an item of a ParseNameBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               name (str): The desired name.
        """

        def __init__(self, name: str,  ID="unassigned"):
            """Constructor
            Args:
                name (str): The desired name.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.name = name

    url = "parseNameBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = ParseNameResponse

    def addItem(self, name: str,  ID="unassigned"):
        """ Adds a ParseNameBatchItem, with the required input, to the batch. 
        Args:
            name (str): The desired name.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.ParseNameBatchItem(name, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "name": item.name,
            })

        return items_list


class ParseNameGeoBatch(Batch):
    """
        A class representing a batch of items whose data is used to infer the likely first/last name structure of a name, ex. John Smith or SMITH, John or SMITH; John. Giving a local context improves precision. 
        Attributes:
            url (str): Ending portion of NamsorAPI url to desired section.
            items (list): A list of all batch items added by the user.
            response_type: The type of response a POST request for this batch will return.
    """
    class ParseNameGeoBatchItem(BatchItem):
        """ 
            A class representing an item of a ParseNameGeoBatch object that contains necessary data which is inputted by the user.
            Attributes:
               ID (str): The ID of the item.
               name (str): The desired name.
               country_code (CountryCodes): The desired country code.
        """

        def __init__(self, name: str, country_code: CountryCodes, ID="unassigned"):
            """Constructor
            Args:
                name (str): The desired name.
                country_code (CountryCodes): The desired country code.
                ID (str, optional): The ID of the item. Defaults to "unassigned".
            """
            self.ID = ID
            self.name = name
            self.country_code = country_code.value

    url = "parseNameBatch"
    items = []
    response_type = None

    def __init__(self):
        """Constructor
        """
        self.items = []
        self.response_type = ParseNameResponse

    def addItem(self, name: str, country_code: CountryCodes, ID="unassigned"):
        """ Adds a ParseNameGeoBatchItem, with the required input, to the batch. 
        Args:
            name (str): The desired name
            country_code (CountryCodes): The desired country code.
            ID (str, optional): The ID of the item. Defaults to "unassigned".
        """
        super().addItem(self.ParseNameGeoBatchItem(name, country_code, ID))

    def batch_item_converter(self) -> list:
        """ Converts all the batch's items' data into a list of dictionaries.
        Returns:
            list: A list of dictionaries each containing each batch item's data.
        """
        items_list = []
        for item in self.items:
            items_list.append({
                "id": item.ID,
                "name": item.name,
                "countryIso2": item.country_code,
            })

        return items_list


def list_separator(data: list) -> list:
    """ Function used to transform the batch arrays into blocks to bypass the 100 item limit.
    Args:
        data (list): The list to be separated.
    Returns:
        list: A list of lists each having a maximum of 100 items.
    """

    big_list = []
    total_num = math.ceil(len(data)/100)
    for i in range(total_num):
        big_list.append(data[i*100:(i+1)*100])
    return big_list



import random

import namsorclient


def test_NameParserType():
    """
    An arbitrary collection of data is created and represented
    as a list of tuples with 2 elements each. The 2 items in a tuple 
    are used to create a sample value for the NameParserType key in
    the response from the ParseName endpoints.

    A NameParserTypeWrapper is instantiated with the generated sample data
    and the assertions ensure that the object is created correctly. 
    """
    tests = [(1, 1), (10, 11), (0, 500), (1190, 0)]
    for i in tests:
        choices = [f'LN{i[0]}FN{i[1]}', f'FN{i[0]}LN{i[1]}']
        raw_string = random.choice(choices)

        tester = namsorclient.NameParserTypeWrapper(raw_string)

        assert str(tester) == raw_string
        assert tester.first_name_count == i[0]
        assert tester.last_name_count == i[1]


def test_FirstLastName():
    test = {
        "id": "4w3vb328f9[h",
        "firstName": "Lelouch",
        "lastName": "Lamperouge"
    }

    tester = namsorclient.FirstLastNameWrapper(test)
    assert tester.ID == test["id"]
    assert tester.first_name == test["firstName"]
    assert tester.last_name == test["lastName"]
    assert str(
        tester) == f'First Name: {test["firstName"]} | Last Name: {test["lastName"]}'


def test_ParseNameResponse():
    test = {
        "id": "null",
        "name": "Lelouch Lamperouge",
        "nameParserType": "LN1FN1",
        "nameParserTypeAlt": "null",
        "firstLastName": {
            "id": "null",
            "firstName": "Lamperouge",
            "lastName": "Lelouch"
        },
        "score": 13.941371643351383
    }
    tester = namsorclient.ParseNameResponse(test)
    assert tester.ID == test["id"]
    assert tester.name == test["name"]
    assert tester.name_parser_type == namsorclient.NameParserTypeWrapper(
        test["nameParserType"])
    assert tester.first_last_name == namsorclient.FirstLastNameWrapper(
        test["firstLastName"])
    assert tester.score == test["score"]


def test_DiasporaResponse():
    test = {
        "id": "null",
        "firstName": "Lelouch",
        "lastName": "Lamperouge",
        "score": 4.723012344810805,
        "ethnicityAlt": "Jewish",
        "ethnicity": "French",
        "lifted": False,
        "countryIso2": "US"
    }
    tester = namsorclient.DiasporaResponse(test)
    assert tester.ID == test["id"]
    assert tester.first_name == test["firstName"]
    assert tester.last_name == test["lastName"]
    assert tester.score == test["score"]
    assert tester.ethnicity == test["ethnicity"]
    assert tester.ethnicity_alt == test["ethnicityAlt"]
    assert tester.lifted == test["lifted"]
    assert tester.country == test["countryIso2"]


def test_RaceEthnicityResponse():
    test = {
        "id": "null",
        "firstName": "Lelouch",
        "lastName": "Lamperouge",
        "raceEthnicityAlt": "B_NL",
        "raceEthnicity": "A",
        "score": 1.1935005311130282
    }

    tester = namsorclient.RaceEthnicityResponse(test)
    assert tester.ID == test["id"]
    assert tester.first_name == test["firstName"]
    assert tester.last_name == test["lastName"]
    assert tester.race_ethnicity_alt == test["raceEthnicityAlt"]
    assert tester.race_ethnicity == test["raceEthnicity"]
    assert tester.score == test["score"]


def test_CountryResponse():
    test = {
        "id": "null",
        "name": "Lelouch Lamperouge",
        "score": 4.588447222213136,
        "country": "FR",
        "countryAlt": "CD",
        "region": "Europe",
        "topRegion": "Europe",
        "subRegion": "Western Europe"
    }

    tester = namsorclient.CountryResponse(test)
    assert tester.ID == test["id"]
    assert tester.name == test["name"]
    assert tester.score == test["score"]
    assert tester.country == test["country"]
    assert tester.country_alt == test["countryAlt"]
    assert tester.region == test["region"]
    assert tester.top_region == test["topRegion"]
    assert tester.sub_region == test["subRegion"]


def test_OriginResponse():
    test = {
        "id": "null",
        "firstName": "Lelouch",
        "lastName": "Lamperouge",
        "countryOrigin": "FR",
        "countryOriginAlt": "GB",
        "score": 1.3850776127906477,
        "regionOrigin": "Europe",
        "topRegionOrigin": "Europe",
        "subRegionOrigin": "Western Europe"
    }

    tester = namsorclient.OriginResponse(test)
    assert tester.ID == test["id"]
    assert tester.first_name == test["firstName"]
    assert tester.last_name == test["lastName"]
    assert tester.country_origin == test["countryOrigin"]
    assert tester.country_origin_alt == test["countryOriginAlt"]
    assert tester.score == test["score"]
    assert tester.region_origin == test["regionOrigin"]
    assert tester.top_region_origin == test["topRegionOrigin"]
    assert tester.sub_region_origin == test["subRegionOrigin"]


def test_GenderFullResponse():
    test = {
        "id": "null",
        "name": "Lelouch Lamperouge",
        "likelyGender": "male",
        "genderScale": -1,
        "score": 99,
    }

    tester = namsorclient.GenderFullResponse(test)
    assert tester.ID == test["id"]
    assert tester.name == test["name"]
    assert tester.likely_gender == test["likelyGender"]
    assert tester.gender_scale == test["genderScale"]
    assert tester.score == test["score"]


def test_GenderResponse():
    test = {
        "id": "null",
        "firstName": "Lelouch",
        "lastName": "Lamperouge",
        "likelyGender": "male",
        "genderScale": -1,
        "score": 9.000730451243589,
        "probabilityCalibrated": 0.6253864122577326
    }

    tester = namsorclient.GenderResponse(test)
    assert tester.ID == test["id"]
    assert tester.first_name == test["firstName"]
    assert tester.last_name == test["lastName"]
    assert tester.likely_gender == test["likelyGender"]
    assert tester.gender_scale == test["genderScale"]
    assert tester.score == test["score"]
    assert tester.probability_calibrated == test["probabilityCalibrated"]



from enum import Enum


class CountryCodes(Enum):
    Aland_Islands = "AX"
    Albania = "AL"
    Algeria = "DZ"
    American_Samoa = "AS"
    Andorra = "AD"
    Angola = "AO"
    Anguilla = "AI"
    Antarctica = "AQ"
    Antigua_and_Barbuda = "AG"
    Argentina = "AR"
    Armenia = "AM"
    Aruba = "AW"
    Australia = "AU"
    Austria = "AT"
    Azerbaijan = "AZ"
    Bahamas = "BS"
    Bahrain = "BH"
    Bangladesh = "BD"
    Barbados = "BB"
    Belarus = "BY"
    Belgium = "BE"
    Belize = "BZ"
    Benin = "BJ"
    Bermuda = "BM"
    Bhutan = "BT"
    Bolivia = "BO"
    Bonaire_Sint_Eustatius_and_Saba = "BQ"
    Bosnia_and_Herzegovina = "BA"
    Botswana = "BW"
    Bouvet_Island = "BV"
    Brazil = "BR"
    British_Indian_Ocean_Territory = "IO"
    Brunei_Darussalam = "BN"
    Bulgaria = "BG"
    Burkina_Faso = "BF"
    Burundi = "BI"
    Cambodia = "KH"
    Cameroon = "CM"
    Canada = "CA"
    Cape_Verde = "CV"
    Cayman_Islands = "KY"
    Central_African_Republic = "CF"
    Chad = "TD"
    Chile = "CL"
    China = "CN"
    Christmas_Island = "CX"
    Cocos_Keeling_Islands = "CC"
    Colombia = "CO"
    Comoros = "KM"
    Congo = "CG"
    Congo_The_Democratic_Republic_of = "CD"
    Cook_Islands = "CK"
    Costa_Rica = "CR"
    IvoryCoast = "CI"
    Croatia = "HR"
    Cuba = "CU"
    Curaçao = "CW"
    Cyprus = "CY"
    Czechia = "CZ"
    Denmark = "DK"
    Djibouti = "DJ"
    Dominica = "DM"
    Dominican_Republic = "DO"
    Ecuador = "EC"
    Egypt = "EG"
    El_Salvador = "SV"
    Equatorial_Guinea = "GQ"
    Eritrea = "ER"
    Estonia = "EE"
    Ethiopia = "ET"
    Falkland_Islands_Malvinas = "FK"
    Faroe_Islands = "FO"
    Fiji = "FJ"
    Finland = "FI"
    France = "FR"
    French_Guiana = "GF"
    French_Polynesia = "PF"
    French_Southern_Territories = "TF"
    Gabon = "GA"
    Gambia = "GM"
    Georgia = "GE"
    Germany = "DE"
    Ghana = "GH"
    Gibraltar = "GI"
    Greece = "GR"
    Greenland = "GL"
    Grenada = "GD"
    Guadeloupe = "GP"
    Guam = "GU"
    Guatemala = "GT"
    Guernsey = "GG"
    Guinea = "GN"
    Guinea_Bissau = "GW"
    Guyana = "GY"
    Haiti = "HT"
    Heard_and_Mc_Donald_Islands = "HM"
    Holy_See_Vatican_City_State = "VA"
    Honduras = "HN"
    Hong_Kong = "HK"
    Hungary = "HU"
    Iceland = "IS"
    India = "IN"
    Indonesia = "ID"
    Iran_Islamic_Republic_of = "IR"
    Iraq = "IQ"
    Ireland = "IE"
    Isle_of_Man = "IM"
    Israel = "IL"
    Italy = "IT"
    Jamaica = "JM"
    Japan = "JP"
    Jersey = "JE"
    Jordan = "JO"
    Kazakstan = "KZ"
    Kenya = "KE"
    Kiribati = "KI"
    Korea_Democratic_Peoples_Republic_of = "KP"
    Korea_Republic_of = "KR"
    Kosovo = "XK"
    Kuwait = "KW"
    Kyrgyzstan = "KG"
    Lao_Peoples_Democratic_Republic = "LA"
    Latvia = "LV"
    Lebanon = "LB"
    Lesotho = "LS"
    Liberia = "LR"
    Libyan_Arab_Jamahiriya = "LY"
    Liechtenstein = "LI"
    Lithuania = "LT"
    Luxembourg = "LU"
    Macao = "MO"
    Macedonia_The_Former_Yugoslav_Republic_Of = "MK"
    Madagascar = "MG"
    Malawi = "MW"
    Malaysia = "MY"
    Maldives = "MV"
    Mali = "ML"
    Malta = "MT"
    Marshall_Islands = "MH"
    Martinique = "MQ"
    Mauritania = "MR"
    Mauritius = "MU"
    Mayotte = "YT"
    Mexico = "MX"
    Micronesia_Federated_States_of = "FM"
    Moldova_Republic_of = "MD"
    Monaco = "MC"
    Mongolia = "MN"
    Montenegro = "ME"
    Montserrat = "MS"
    Morocco = "MA"
    Mozambique = "MZ"
    Myanmar = "MM"
    Namibia = "NA"
    Nauru = "NR"
    Nepal = "NP"
    Netherlands = "NL"
    Netherlands_Antilles = "AN"
    New_Caledonia = "NC"
    New_Zealand = "NZ"
    Nicaragua = "NI"
    Niger = "NE"
    Nigeria = "NG"
    Niue = "NU"
    Norfolk_Island = "NF"
    Northern_Mariana_Islands = "MP"
    Norway = "NO"
    Oman = "OM"
    Pakistan = "PK"
    Palau = "PW"
    Palestinian_Territory_Occupied = "PS"
    Panama = "PA"
    Papua_New_Guinea = "PG"
    Paraguay = "PY"
    Peru = "PE"
    Philippines = "PH"
    Pitcairn = "PN"
    Poland = "PL"
    Portugal = "PT"
    Puerto_Rico = "PR"
    Qatar = "QA"
    Republic_of_Serbia = "RS"
    Reunion = "RE"
    Romania = "RO"
    Russia_Federation = "RU"
    Rwanda = "RW"
    Saint_Barthélemy = "BL"
    Saint_Helena = "SH"
    Saint_Kitts_and_Nevis = "KN"
    Saint_Lucia = "LC"
    Saint_Martin = "MF"
    Saint_Pierre_and_Miquelon = "PM"
    Saint_Vincent_and_the_Grenadines = "VC"
    Samoa = "WS"
    San_Marino = "SM"
    Sao_Tome_and_Principe = "ST"
    Saudi_Arabia = "SA"
    Senegal = "SN"
    Serbia_and_Montenegro = "CS"
    Seychelles = "SC"
    Sierra_Leone = "SL"
    Singapore = "SG"
    Sint_Maarten = "SX"
    Slovakia = "SK"
    Slovenia = "SI"
    Solomon_Islands = "SB"
    Somalia = "SO"
    South_Africa = "ZA"
    South_Georgia_and_The_South_Sandwich_Islands = "GS"
    South_Sudan = "SS"
    Spain = "ES"
    Sri_Lanka = "LK"
    Sudan = "SD"
    Suriname = "SR"
    Svalbard_and_Jan_Mayen = "SJ"
    Swaziland = "SZ"
    Sweden = "SE"
    Switzerland = "CH"
    Syrian_Arab_Republic = "SY"
    Taiwan_Province_of_China = "TW"
    Tajikistan = "TJ"
    Tanzania_United_Republic_of = "TZ"
    Thailand = "TH"
    Timor_Leste = "TL"
    Togo = "TG"
    Tokelau = "TK"
    Tonga = "TO"
    Trinidad_and_Tobago = "TT"
    Tunisia = "TN"
    Turkey = "TR"
    Turkish_Rep_N_Cyprus = "XT"
    Turkmenistan = "TM"
    Turks_and_Caicos_Islands = "TC"
    Tuvalu = "TV"
    Uganda = "UG"
    Ukraine = "UA"
    United_Arab_Emirates = "AE"
    United_Kingdom = "GB"
    United_States = "US"
    United_States_Minor_Outlying_Islands = "UM"
    Uruguay = "UY"
    Uzbekistan = "UZ"
    Vanuatu = "VU"
    Venezuela = "VE"
    Vietnam = "VN"
    Virgin_Islands_British = "VG"
    Virgin_Islands_US = "VI"
    Wallis_and_Futuna = "WF"
    Western_Sahara = "EH"
    Yemen = "YE"
    Zambia = "ZM"
    Zimbabwe = "ZW"


def create(value: str) -> CountryCodes:
    for country in CountryCodes:
        if country.value == value:
            return country



import json
import requests
import time

from namsorclient.models import *
from namsorclient.request_objects import *
from namsorclient.country_codes import CountryCodes


BASE_URL = "https://v2.namsor.com/NamSorAPIv2/api2/json/"


class NamsorClient:
    """
    A class representing a NamsorAPI client
    Attributes:
        api_key(str): The Namsor API key of the user.
    """

    def __init__(self, api_key):
        self.api_key = api_key
        test_response = self.__api_get()
        if test_response.status_code == 401:
            raise Exception('Invalid API Key')
            # The client has not entered his/her API key or the API key is incorrect
        elif test_response.status_code == 403:
            raise Exception("API Limit Reached or API Key Disabled")
            # The client's amount of requests has reached the maximum amount he/she can send in a month or his/her subscription plan has been cancelled

    def __api_get(self, url="gender/Lelouch/Lamperouge") -> requests.models.Response:
        """ returns a response containing desired information of the data
        Args:
            url (str, optional): ending portion of NamsorAPI url to desired section. Defaults to "gender/Lelouch/Lamperouge".
        Returns:
            requests.models.Response: the response received from this GET request
        """

        # While the choice to include the error checking in the get request wrapper does somewhat compromise the elegance of the code base, the alternative, which would mean performing a request prior, potentially wastes a classification from the user's account and thus it was decided to prioritize the user's experience, especially given the non efficiency-intensive nature of the likely utilization of the wrapper

        return requests.get(url=f"{BASE_URL}{url}", headers={"X-API-KEY": self.api_key})

    def gender(self, first_name: str, last_name: str) -> GenderResponse:
        """ Infer the likely gender of a name.
        Args:
            first_name (str): The desired first name. 
            last_name (str): The desired last name.
        Returns:
            GenderResponse: An Object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"gender/{first_name}/{last_name}"
        return GenderResponse(self.__api_get(url=url).json())

    def genderGeo(self, first_name: str, last_name: str, country_code: CountryCodes) -> GenderResponse:
        """Infer the likely gender of a name, given a local context (ISO2 country code).
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
            country_code (CountryCodes): The country code, to be passed using the CountryCodes object.
        Returns:
            GenderResponse: An Object which is a wrapper of the API's response object for this particular endpoint.
        """
        url = f"genderGeo/{first_name}/{last_name}/{country_code.value}"
        return GenderResponse(self.__api_get(url=url).json())

    def genderFullGeo(self, full_name: str, country_code: CountryCodes) -> GenderResponse:
        """Infer the likely gender of a full name, given a local context (ISO2 country code).
        Args:
            full_name (str): The name to be classified.
            country_code (CountryCodes): The country code to aid with classification.
        Returns:
            GenderResponse: An Object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"genderFullGeo/{full_name}/{country_code.value}"
        return GenderResponse(self.__api_get(url=url).json())

    def genderFull(self, full_name: str) -> GenderResponse:
        """Infer the likely gender of a full name, ex. John H. Smith
        Args:
            full_name (str): The name to be classified.
        Returns:
            GenderResponse: An Object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"genderFull/{full_name}"
        return GenderResponse(self.__api_get(url=url).json())

    def usRaceEthnicity(self, first_name: str, last_name: str) -> RaceEthnicityResponse:
        """Infer a US resident's likely race/ethnicity according to US Census taxonomy W_NL (white, non latino), HL (hispano latino),  A (asian, non latino), B_NL (black, non latino).
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name.
        Returns:
            OriginResponse: An object which is a wrapper of the API's response object for this particular endpoint.
        """
        url = f"usRaceEthnicity/{first_name}/{last_name}"
        return RaceEthnicityResponse(self.__api_get(url=url).json())

    def usRaceEthnicityZIP5(self, first_name: str, last_name: str, zip5_code: str) -> RaceEthnicityResponse:
        """Infer a US resident's likely race/ethnicity according to US Census taxonomy, using ZIP5 code info. Output is W_NL (white, non latino), HL (hispano latino),  A (asian, non latino), B_NL (black, non latino).
        Args:
            first_name (str): The desired first name
            last_name (str): The desired last name
            zip5_code (str): The zip code to aid with classification
        Returns:
            OriginResponse: An object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"usRaceEthnicity/{first_name}/{last_name}/{zip5_code}"
        return RaceEthnicityResponse(self.__api_get(url=url).json())

    def diaspora(self, first_name: str, last_name: str, country_code: CountryCodes) -> DiasporaResponse:
        """ Infer the likely ethnicity/diaspora of a personal name, given a country of residence ISO2 code
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name
            country_code (CountryCodes): The country code to aid with classification.
        Returns:
            OriginResponse: An object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"diaspora/{country_code.value}/{first_name}/{last_name}"
        return DiasporaResponse(self.__api_get(url=url).json())

    def parseName(self, full_name: str) -> ParseNameResponse:
        """Infer the likely first/last name structure of a name, ex. John Smith or SMITH, John or SMITH; John.
        Args:
            full_name (str): The name to be parsed
        Returns: 
        ParseNameResponse: An object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"parseName/{full_name}"
        return ParseNameResponse(self.__api_get(url=url).json())

    def parseNameGeo(self, full_name: str, country_code: CountryCodes) -> ParseNameResponse:
        """Infer the likely first/last name structure of a name, ex. John Smith or SMITH, John or SMITH; John, given an ISO2 country of residence.
        Args:
            full_name (str): The full name to be parsed.
            coutry_code (CountryCodes): The country code to aid with parsing
        Returns:
            ParseNameResponse: An object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"parseName/{full_name}/{country_code.value}"
        return ParseNameResponse(self.__api_get(url=url).json())

    def origin(self, first_name: str, last_name: str) -> OriginResponse:
        """Infer the likely country of origin of a personal name. Assumes names as they are in the country of origin. For US, CA, AU, NZ and other melting-pots : use 'diaspora' instead.
        Args:
            first_name (str): The desired first name.
            last_name (str): The desired last name
        Returns:
            OriginResponse: An object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"origin/{first_name}/{last_name}"
        return OriginResponse(self.__api_get(url=url).json())

    def country(self, full_name: str) -> CountryResponse:
        """Infer the likely country of residence of a personal full name, or one surname. Assumes names as they are in the country of residence OR the country of origin.
        Args:
            full_name (str): The name whose country of residence should be determined.
        Returns:
            OriginResponse: An object which is a wrapper of the API's response object for this particular endpoint.
        """

        url = f"country/{first_name}/{last_name}"
        return OriginResponse(self.__api_get(url=url).json())
